@@ -1,6 +1,6 @@
 /*
  * The initial "port" of dwm to curses was done by
- * (c) 2007-2011 Marc Andre Tanner <mat at brain-dump dot org>
+ * (c) 2007-2012 Marc Andre Tanner <mat at brain-dump dot org>
  *
  * It is highly inspired by the original X11 dwm and
  * reuses some code of it which is mostly
@@ -39,6 +39,10 @@
 int ESCDELAY;
 #endif
 
+#ifndef NCURSES_REENTRANT
+# define set_escdelay(d) (ESCDELAY = (d))
+#endif
+
 typedef struct {
 	double mfact;
 	int history;
@@ -80,6 +84,9 @@ typedef struct {
 } ColorRule;
 
 #define ALT(k)      ((k) + (161 - 'a'))
+#if defined CTRL && defined _AIX
+  #undef CTRL
+#endif
 #ifndef CTRL
   #define CTRL(k)   ((k) & 0x1F)
 #endif
@@ -248,9 +255,10 @@ drawbar() {
 
 static void
 draw_border(Client *c) {
-	char *s, t = '\0';
-	int x, y, o;
-	if (sel == c) {
+	char t = '\0';
+	int x, y, maxlen;
+
+	if (sel == c || (runinall && !c->minimized)) {
 		wattrset(c->window, SELECTED_ATTR);
 		wcolor_set(c->window, vt_color_get(SELECTED_FG, SELECTED_BG), NULL);
 	} else {
@@ -260,19 +268,20 @@ draw_border(Client *c) {
 	getyx(c->window, y, x);
 	curs_set(0);
 	mvwhline(c->window, 0, 0, ACS_HLINE, c->w);
-	o = c->w - (4 + sstrlen(TITLE) - 5  + sstrlen(SEPARATOR));
-	if (o < 0)
-		o = 0;
-	if ((size_t)o < sizeof(c->title)) {
-		t = *(s = &c->title[o]);
-		*s = '\0';
+	maxlen = c->w - (2 + sstrlen(TITLE) - sstrlen("%s%sd")  + sstrlen(SEPARATOR) + 2);
+	if (maxlen < 0)
+		maxlen = 0;
+	if ((size_t)maxlen < sizeof(c->title)) {
+		t = c->title[maxlen];
+		c->title[maxlen] = '\0';
 	}
+
 	mvwprintw(c->window, 0, 2, TITLE,
 	          *c->title ? c->title : "",
 	          *c->title ? SEPARATOR : "",
 	          c->order);
 	if (t)
-		*s = t;
+		c->title[maxlen] = t;
 	wmove(c->window, y, x);
 	if (!c->minimized)
 		curs_set(vt_cursor(c->term));
@@ -393,7 +402,7 @@ focus(Client *c) {
 }
 
 static void
-applycolorrules(Vt *term, char *title) {
+applycolorrules(Client *c) {
 	unsigned int i;
 	unsigned attrs = A_NORMAL;
 	short fg = -1, bg = -1;
@@ -401,31 +410,29 @@ applycolorrules(Vt *term, char *title) {
 
 	for (i = 0; i < countof(colorrules); i++) {
 		r = &colorrules[i];
-		if (strstr(title, r->title)) {
+		if (strstr(c->title, r->title)) {
 			attrs = r->attrs;
 			fg = r->fg;
 			bg = r->bg;
 			break;
 		}
 	}
-	vt_set_default_colors(term, attrs, fg, bg);
+
+	vt_set_default_colors(c->term, attrs, fg, bg);
 }
 
-static int
-title_escape_seq_handler(Vt *term, char *es) {
-	Client *c;
-	unsigned int l;
-	if (es[0] != ']' || (es[1] && (es[1] < '0' || es[1] > '9')) || (es[2] && es[2] != ';'))
-		return VT_HANDLER_NOWAY;
-	if ((l = strlen(es)) < 3 || es[l - 1] != '\07')
-		return VT_HANDLER_NOTYET;
-	es[l - 1] = '\0';
-	c = (Client *)vt_get_data(term);
-	strncpy(c->title, es + 3, sizeof(c->title));
-	draw_border(c);
-	debug("window title: %s\n", c->title);
-	applycolorrules(term, c->title);
-	return VT_HANDLER_OK;
+static void
+term_event_handler(Vt *term, int event, void *event_data) {
+	Client *c = (Client *)vt_get_data(term);
+	switch (event) {
+	case VT_EVENT_TITLE:
+		if (event_data)
+			strncpy(c->title, event_data, sizeof(c->title) - 1);
+		c->title[event_data ? sizeof(c->title) - 1 : 0] = '\0';
+		draw_border(c);
+		applycolorrules(c);
+		break;
+	}
 }
 
 static void
@@ -703,11 +710,13 @@ create(const char *args[]) {
 	c->window = newwin(wah, waw, way, wax);
 	c->term = vt_create(screen.h - 1, screen.w, screen.history);
 	c->cmd = cmd;
-	if (args && args[1])
-		strncpy(c->title, args[1], sizeof(c->title));
+	if (args && args[1]) {
+		strncpy(c->title, args[1], sizeof(c->title) - 1);
+		c->title[sizeof(c->title) - 1] = '\0';
+	}
 	c->pid = vt_forkpty(c->term, "/bin/sh", pargs, env, &c->pty);
 	vt_set_data(c->term, c);
-	vt_set_handler(c->term, title_escape_seq_handler);
+	vt_set_event_handler(c->term, term_event_handler);
 	c->w = screen.w;
 	c->h = screen.h;
 	c->x = wax;
@@ -975,6 +984,7 @@ togglemouse(const char *args[]) {
 static void
 togglerunall(const char *args[]) {
 	runinall = !runinall;
+	draw_all(true);
 }
 
 static void
@@ -1214,7 +1224,7 @@ parse_args(int argc, char *argv[]) {
 	bool init = false;
 
 	if (!getenv("ESCDELAY"))
-		ESCDELAY = 100;
+		set_escdelay(100);
 	for (arg = 1; arg < argc; arg++) {
 		if (argv[arg][0] != '-') {
 			const char *args[] = { argv[arg], NULL };
@@ -1229,7 +1239,7 @@ parse_args(int argc, char *argv[]) {
 			usage();
 		switch (argv[arg][1]) {
 			case 'v':
-				puts("dvtm-"VERSION" (c) 2007-2011 Marc Andre Tanner");
+				puts("dvtm-"VERSION" (c) 2007-2012 Marc Andre Tanner");
 				exit(EXIT_SUCCESS);
 			case 'm': {
 				char *mod = argv[++arg];
@@ -1240,11 +1250,11 @@ parse_args(int argc, char *argv[]) {
 				break;
 			}
 			case 'd':
-				ESCDELAY = atoi(argv[++arg]);
+				set_escdelay(atoi(argv[++arg]));
 				if (ESCDELAY < 50)
-					ESCDELAY = 50;
+					set_escdelay(50);
 				else if (ESCDELAY > 1000)
-					ESCDELAY = 1000;
+					set_escdelay(1000);
 				break;
 			case 'h':
 				screen.history = atoi(argv[++arg]);
