@@ -1,7 +1,7 @@
 /*
  * Copyright © 2004 Bruno T. C. de Oliveira
  * Copyright © 2006 Pierre Habouzit
- * Copyright © 2008-2011 Marc Andre Tanner
+ * Copyright © 2008-2012 Marc Andre Tanner
  *
  * Permission to use, copy, modify, and distribute this software for any
  * purpose with or without fee is hereby granted, provided that the above
@@ -29,7 +29,7 @@
 #include <sys/types.h>
 #include <termios.h>
 #include <wchar.h>
-#ifdef __linux__
+#if defined(__linux__) || defined(__CYGWIN__)
 # include <pty.h>
 #elif defined(__FreeBSD__)
 # include <libutil.h>
@@ -42,6 +42,10 @@
 
 #include "vt.h"
 
+#ifdef _AIX
+# include "forkpty-aix.c"
+#endif
+
 #ifndef NCURSES_ATTR_SHIFT
 # define NCURSES_ATTR_SHIFT 8
 #endif
@@ -58,6 +62,7 @@
 #define MIN(x, y) ((x) < (y) ? (x) : (y))
 #define COLOR_PALETTE_START 1
 #define COLOR_PALETTE_END (unsigned)(MIN(512, COLOR_PAIRS))
+#define sstrlen(str) (sizeof(str) - 1)
 
 static bool is_utf8, has_default_colors, use_color_palette;
 static short *color2palette;
@@ -130,10 +135,15 @@ struct Vt {
 	mbstate_t ps;
 	char rbuf[BUFSIZ];
 	char ebuf[BUFSIZ];
-	int rlen, elen;
+	unsigned int rlen, elen;
+
+	/* xterm style window title */
+	char title[256];
+
+	vt_event_handler_t event_handler;
 
 	/* custom escape sequence handler */
-	vt_handler_t handler;
+	vt_escseq_handler_t escseq_handler;
 	void *data;
 };
 
@@ -190,6 +200,7 @@ static char const * const keytable[KEY_MAX+1] = {
 	[KEY_F(18)]     = "\e[32~",
 	[KEY_F(19)]     = "\e[33~",
 	[KEY_F(20)]     = "\e[34~",
+	[KEY_RESIZE]    = "",
 };
 
 static void process_nonprinting(Vt *t, wchar_t wc);
@@ -237,10 +248,13 @@ static void row_roll(Row *start, Row *end, int count)
 
 static void clamp_cursor_to_bounds(Vt *t)
 {
-	if (t->curs_row < t->lines)
-		t->curs_row = t->lines;
-	if (t->curs_row >= t->lines + t->rows)
-		t->curs_row = t->lines + t->rows - 1;
+	Row *lines = t->relposmode ? t->scroll_top : t->lines;
+	int rows = t->relposmode ? t->scroll_bot - t->scroll_top : t->rows;
+
+	if (t->curs_row < lines)
+		t->curs_row = lines;
+	if (t->curs_row >= lines + rows)
+		t->curs_row = lines + rows - 1;
 	if (t->curs_col < 0)
 		t->curs_col = 0;
 	if (t->curs_col >= t->cols)
@@ -295,7 +309,7 @@ static void fill_scroll_buf(Vt *t, int s)
 
 	if (s > 0 && t->scroll_buf_sz) {
 		for (int i = 0; i < s; i++) {
-			struct Row tmp = t->scroll_top[i];
+			Row tmp = t->scroll_top[i];
 			t->scroll_top[i] = t->scroll_buf[t->scroll_buf_ptr];
 			t->scroll_buf[t->scroll_buf_ptr] = tmp;
 
@@ -311,7 +325,7 @@ static void fill_scroll_buf(Vt *t, int s)
 			if (t->scroll_buf_ptr == -1)
 				t->scroll_buf_ptr = t->scroll_buf_sz - 1;
 
-			struct Row tmp = t->scroll_top[i];
+			Row tmp = t->scroll_top[i];
 			t->scroll_top[i] = t->scroll_buf[t->scroll_buf_ptr];
 			t->scroll_buf[t->scroll_buf_ptr] = tmp;
 			t->scroll_top[i].dirty = true;
@@ -467,14 +481,16 @@ static void interpret_csi_ED(Vt *t, int param[], int pcount)
 /* interprets a 'move cursor' (CUP) escape sequence */
 static void interpret_csi_CUP(Vt *t, int param[], int pcount)
 {
+	Row *lines = t->relposmode ? t->scroll_top : t->lines;
+
 	if (pcount == 0) {
-		t->curs_row = (t->relposmode ? t->scroll_top : t->lines);
+		t->curs_row = lines;
 		t->curs_col = 0;
 	} else if (pcount == 1) {
-		t->curs_row = (t->relposmode ? t->scroll_top : t->lines) + param[0] - 1;
+		t->curs_row = lines + param[0] - 1;
 		t->curs_col = 0;
 	} else {
-		t->curs_row = (t->relposmode ? t->scroll_top : t->lines) + param[0] - 1;
+		t->curs_row = lines + param[0] - 1;
 		t->curs_col = param[1] - 1;
 	}
 
@@ -611,7 +627,7 @@ static void interpret_csi_ECH(Vt *t, int param[], int pcount)
 {
 	int n = (pcount && param[0] > 0) ? param[0] : 1;
 
-	if (t->curs_col + n < t->cols)
+	if (t->curs_col + n > t->cols)
 		n = t->cols - t->curs_col;
 
 	row_set(t->curs_row, t->curs_col, n, t);
@@ -822,6 +838,24 @@ static void interpret_esc_SCS(Vt *t)
 	t->graphmode = t->charsets[0];
 }
 
+/* Interpret xterm specific escape sequences */
+static void interpret_esc_xterm(Vt *t)
+{
+	/* ESC]n;dataBEL -- the ESC is not part of t->ebuf */
+	char *title = NULL;
+
+	switch (t->ebuf[1]) {
+	case '0':
+	case '2':
+		t->ebuf[t->elen - 1] = '\0';
+		if (t->elen > sstrlen("]n;\a"))
+			title = t->ebuf + sstrlen("]n;");
+
+		if (t->event_handler)
+			t->event_handler(t, VT_EVENT_TITLE, title);
+	}
+}
+
 static void try_interpret_escape_seq(Vt *t)
 {
 	char lastchar = t->ebuf[t->elen - 1];
@@ -829,18 +863,25 @@ static void try_interpret_escape_seq(Vt *t)
 	if (!*t->ebuf)
 		return;
 
-	if (t->handler) {
-		switch ((*(t->handler)) (t, t->ebuf)) {
-		case VT_HANDLER_OK:
-			goto cancel;
-		case VT_HANDLER_NOTYET:
-			if (t->elen + 1 >= (int)sizeof(t->ebuf))
+	if (t->escseq_handler) {
+		switch ((*(t->escseq_handler)) (t, t->ebuf)) {
+		case VT_ESCSEQ_HANDLER_OK:
+			cancel_escape_sequence(t);
+			return;
+		case VT_ESCSEQ_HANDLER_NOTYET:
+			if (t->elen + 1 >= sizeof(t->ebuf))
 				goto cancel;
 			return;
 		}
 	}
 
 	switch (*t->ebuf) {
+	case '#': /* ignore DECDHL, DECSWL, DECDWL, DECHCP, DECALN, DECFPP */
+		if (t->elen == 2) {
+			cancel_escape_sequence(t);
+			return;
+		}
+		break;
 	case '(':
 	case ')':
 		if (t->elen == 2) {
@@ -850,8 +891,11 @@ static void try_interpret_escape_seq(Vt *t)
 		}
 		break;
 	case ']': /* xterm thing */
-		if (lastchar == '\a')
-			goto cancel;
+		if (lastchar == '\a') {
+			interpret_esc_xterm(t);
+			cancel_escape_sequence(t);
+			return;
+		}
 		break;
 	case '[':
 		if (is_valid_csi_ender(lastchar)) {
@@ -886,11 +930,11 @@ static void try_interpret_escape_seq(Vt *t)
 		goto cancel;
 	}
 
-	if (t->elen + 1 >= (int)sizeof(t->ebuf)) {
+	if (t->elen + 1 >= sizeof(t->ebuf)) {
 cancel:
 #ifndef NDEBUG
 		fprintf(stderr, "cancelled: \\033");
-		for (int i = 0; i < (int)t->elen; i++) {
+		for (unsigned int i = 0; i < t->elen; i++) {
 			if (isprint(t->ebuf[i])) {
 				fputc(t->ebuf[i], stderr);
 			} else {
@@ -981,7 +1025,7 @@ static void put_wc(Vt *t, wchar_t wc)
 	}
 
 	if (t->escaped) {
-		if (t->elen + 1 < (int)sizeof(t->ebuf)) {
+		if (t->elen + 1 < sizeof(t->ebuf)) {
 			t->ebuf[t->elen] = wc;
 			t->ebuf[++t->elen] = '\0';
 			try_interpret_escape_seq(t);
@@ -1050,7 +1094,8 @@ static void put_wc(Vt *t, wchar_t wc)
 
 int vt_process(Vt *t)
 {
-	int res, pos = 0;
+	int res;
+	unsigned int pos = 0;
 
 	if (t->pty < 0) {
 		errno = EINVAL;
@@ -1549,9 +1594,14 @@ void vt_init(void)
 	is_utf8_locale();
 }
 
-void vt_set_handler(Vt *t, vt_handler_t handler)
+void vt_set_escseq_handler(Vt *t, vt_escseq_handler_t handler)
 {
-	t->handler = handler;
+	t->escseq_handler = handler;
+}
+
+void vt_set_event_handler(Vt *t, vt_event_handler_t handler)
+{
+	t->event_handler = handler;
 }
 
 void vt_set_data(Vt *t, void *data)
